use std::{error::Error, sync::Arc};

use vulkano::{
    buffer::Subbuffer,
    command_buffer::{
        AutoCommandBufferBuilder, PrimaryAutoCommandBuffer, RenderPassBeginInfo, SubpassBeginInfo,
        SubpassEndInfo,
    },
    descriptor_set::allocator::StandardDescriptorSetAllocator,
    pipeline::GraphicsPipeline,
    render_pass::{Framebuffer, RenderPass},
};

pub trait AsRecordable<T> {
    fn record_render_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        descriptor_set_allocator: &Arc<StandardDescriptorSetAllocator>,
        pipeline: &Arc<GraphicsPipeline>,
        data: &T,
    ) -> Result<(), Box<dyn Error>>;
}

pub trait AsRenderableMesh<V, I> {
    fn vertex_buffer(&self) -> &Subbuffer<[V]>;

    fn vertex_count(&self) -> u32;

    fn first_vertex(&self) -> u32 {
        0
    }

    fn vertex_offset(&self) -> i32 {
        0
    }

    fn index_buffer(&self) -> Option<&Subbuffer<[I]>> {
        None
    }

    fn index_count(&self) -> u32 {
        0
    }

    fn first_index(&self) -> u32 {
        0
    }

    fn index_offset(&self) -> i32 {
        0
    }
}

pub trait AsRenderPassRecordable<T> {
    fn begin_render_pass(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        _render_pass: &Arc<RenderPass>,
        framebuffer: &Arc<Framebuffer>,
        clear_values: Vec<Option<vulkano::format::ClearValue>>,
    ) -> Result<(), Box<dyn Error>> {
        builder.begin_render_pass(
            RenderPassBeginInfo {
                clear_values,
                ..RenderPassBeginInfo::framebuffer(framebuffer.clone())
            },
            SubpassBeginInfo {
                contents: vulkano::command_buffer::SubpassContents::Inline,
                ..Default::default()
            },
        )?;
        Ok(())
    }

    fn end_render_pass(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) -> Result<(), Box<dyn Error>> {
        builder.end_render_pass(SubpassEndInfo::default())?;
        Ok(())
    }

    fn record_render_pass_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        data: &T,
    ) -> Result<(), Box<dyn Error>>;
}
